@@ -6,7 +6,8 @@ from tensorflow.python.framework import tensor_shape
 
 import tensorflow_probability as tfp
 
-from b4tf.utils import ReciprocalGammaInitializer
+from b4tf.utils import (ReciprocalGammaInitializer,
+                        safe_div, safe_exp, non_negative_constraint)
 
 class PBPLayer(tf.keras.layers.Layer):
     """
@@ -40,7 +41,7 @@ class PBPLayer(tf.keras.layers.Layer):
         self.inv_V1 = tf.math.square(self.inv_sqrtV1)
 
 
-        over_gamma = ReciprocalGammaInitializer(1.0,1e-6)
+        over_gamma = ReciprocalGammaInitializer(6.0,6.0)
         self.kernel_m = self.add_weight("kernel_mean",
                                         shape=[last_dim,self.units],
                                         initializer=tf.keras.initializers.HeNormal(),
@@ -82,29 +83,25 @@ class PBPLayer(tf.keras.layers.Layer):
 
         # Kernel
         self.kernel_m.assign_add(self.kernel_v * dlogZ_dkm)
-        new_kv = (self.kernel_v -
-                  (tf.math.square(self.kernel_v) *
-                   (tf.math.square(dlogZ_dkm) - 2*dlogZ_dkv)))
-        self.kernel_v.assign(tf.math.maximum(new_kv,tf.zeros_like(new_kv)))
+        new_kv = self.kernel_v - (tf.math.square(self.kernel_v) *
+                                  (tf.math.square(dlogZ_dkm) - 2*dlogZ_dkv))
+        self.kernel_v.assign(non_negative_constraint(new_kv))
 
         # Bias
         self.bias_m.assign_add(self.bias_v * dlogZ_dbm)
-        new_bv = (self.bias_v -
-                  (tf.math.square(self.bias_v) *
-                   (tf.math.square(dlogZ_dbm) - 2*dlogZ_dbv)))
-        self.bias_v.assign(tf.math.maximum(new_bv,tf.zeros_like(new_bv)))
+        new_bv = self.bias_v - (tf.math.square(self.bias_v) *
+                                (tf.math.square(dlogZ_dbm) - 2*dlogZ_dbv))
+        self.bias_v.assign(non_negative_constraint(new_bv))
 
     @tf.function
     def _sample_weights(self):
-        eps = self.Normal.sample(self.kernel_m.shape)
-        std = tf.math.sqrt(tf.math.maximum(self.kernel_v,
-                                           tf.zeros_like(self.kernel_v)))
-        W = self.kernel_m + std * eps
+        eps_k = self.Normal.sample(self.kernel_m.shape)
+        std_k = tf.math.sqrt(tf.maximum(self.kernel_v, tf.zeros_like(self.kernel_v)))
+        W = self.kernel_m + std_k * eps_k
 
-        eps = self.Normal.sample(self.bias_m.shape)
-        std = tf.math.sqrt(tf.math.maximum(self.bias_v,
-                                           tf.zeros_like(self.bias_v)))
-        b = self.bias_m + std * eps
+        eps_b = self.Normal.sample(self.bias_m.shape)
+        std_b = tf.math.sqrt(tf.maximum(self.bias_v, tf.zeros_like(self.bias_v)))
+        b = self.bias_m + std_b * eps_b
         return W, b
 
     @tf.function
@@ -202,13 +199,13 @@ class PBPReLULayer(PBPLayer):
         """
         ma, va = super().predict(m_prev,v_prev)
 
-        _sqrt_v = tf.math.sqrt(tf.math.maximum(va,tf.zeros_like(va))) + 1e-6
-        _alpha = ma / _sqrt_v
-        _inv_alpha = 1.0/(_alpha + 1e-6)
+        _sqrt_v = tf.math.sqrt(tf.maximum(va,tf.zeros_like(va)))
+        _alpha = safe_div(ma, _sqrt_v)
+        _inv_alpha = safe_div(tf.constant(1.0,dtype=_alpha.dtype), _alpha)
         _cdf_alpha = self.Normal.cdf(_alpha)
         _gamma = tf.where(_alpha < -30,
                           -_alpha + _inv_alpha * (-1 + 2*tf.math.square(_inv_alpha)),
-                          self.Normal.prob(-_alpha)/_cdf_alpha)
+                          safe_div(self.Normal.prob(-_alpha), _cdf_alpha))
         _vp = ma + _sqrt_v * _gamma
 
         mb = _cdf_alpha * _vp
@@ -243,8 +240,8 @@ class PBP:
             Data type
         """
         self.dtype = tf.as_dtype(dtype)
-        self.alpha = tf.Variable(1.0,trainable=True,dtype=self.dtype)
-        self.beta  = tf.Variable(0.0,trainable=True,dtype=self.dtype)
+        self.alpha = tf.Variable(6.0,trainable=True,dtype=self.dtype)
+        self.beta  = tf.Variable(6.0,trainable=True,dtype=self.dtype)
 
         pi = tf.math.atan(tf.constant(1.0,dtype=self.dtype)) * 4
         self.log_inv_sqrt2pi = -0.5*tf.math.log(2.0*pi)
@@ -269,15 +266,47 @@ class PBP:
             l.build(last_shape)
             self.layers.append(l)
 
+        self.trainables = [l.trainable_weights for l in self.layers]
 
         self.Normal=tfp.distributions.Normal(loc=tf.constant(0.0,dtype=self.dtype),
                                              scale=tf.constant(1.0,dtype=self.dtype))
         self.Gamma = tfp.distributions.Gamma(concentration=self.alpha,
                                              rate=self.beta)
 
+    @tf.function
     def _logZ(self,diff_square: tf.Tensor, v: tf.Tensor):
-        return tf.reduce_sum(-0.5 * (diff_square / (v+1e-6)) +
-                             self.log_inv_sqrt2pi - tf.math.log(v))
+        v0 = v + 1e-6
+        return tf.reduce_sum(-0.5 * (diff_square / v0) +
+                             self.log_inv_sqrt2pi - 0.5*tf.math.log(v0))
+
+    @tf.function
+    def _logZ1_minus_logZ2(self,diff_square: tf.Tensor, v1: tf.Tensor, v2: tf.Tensor):
+        """
+        log Z1 - log Z2
+
+        Parameters
+        ----------
+        diff_square : tf.Tensor
+            (y - m)^2
+        v1 : tf.Tensor
+            Z1 = Z(diff_squzre,v1)
+        v2 : tf.Tensor
+            Z2 = Z(diff_square,v2)
+
+        Returns
+        -------
+        : tf.Tensor
+            log Z1 - log Z2
+
+
+        Notes
+        -----
+        diff_square >= 0
+        v1 >= 0
+        v2 >= 0
+        """
+        return tf.reduce_sum(-0.5 * diff_square * safe_div(v2-v1, v1*v2)
+                             -0.5 * tf.math.log(safe_div(v1, v2) + 1e-6))
 
 
     def _ensure_input(self,x):
@@ -319,7 +348,7 @@ class PBP:
         return y
 
 
-    def fit(self,x,y):
+    def fit(self,x,y,batch_size:int = 16):
         """
         Fit posterior distribution with observation
 
@@ -333,43 +362,50 @@ class PBP:
         x = self._ensure_input(x)
         y = self._ensure_output(y)
 
-        self._fit(x,y)
+        data = tf.data.Dataset.from_tensor_slices((x,y)).batch(batch_size)
+
+        for _x,_y in data:
+            self._fit(_x,_y)
 
     @tf.function
     def _fit(self,x: tf.Tensor, y: tf.Tensor):
-        trainables = [l.trainable_weights for l in self.layers]
         with tf.GradientTape() as tape:
-            tape.watch(trainables)
+            tape.watch(self.trainables)
             m, v = self._predict(x)
 
-            v0 = v + self.beta/((self.alpha - 1)+1e-6)
+            v0 = v + safe_div(self.beta, self.alpha - 1)
             diff_square = tf.math.square(y - m)
             logZ0 = self._logZ(diff_square,v0)
 
-        grad = tape.gradient(logZ0,trainables)
+        grad = tape.gradient(logZ0,self.trainables)
         for l, g in zip(self.layers, grad):
             l.apply_gradient(g)
 
 
-        alpha0 = self.alpha + 1e-6
         alpha1 = self.alpha + 1
-        v1 = v + self.beta/alpha0
+        v1 = v + safe_div(self.beta, self.alpha)
         v2 = v + self.beta/alpha1
 
-        logZ1 = self._logZ(diff_square,v1)
-        logZ2 = self._logZ(diff_square,v2)
+        logZ2_logZ1 = self._logZ1_minus_logZ2(diff_square,v1=v2,v2=v1)
+        logZ1_logZ0 = self._logZ1_minus_logZ2(diff_square,v1=v1,v2=v0)
 
-        logZ2_logZ1 = logZ2 - logZ1
-        logZ1_logZ0 = logZ1 - logZ0
+        logZ_diff = logZ2_logZ1 - logZ1_logZ0
+
+        Z0Z2_Z1Z1 = safe_exp(logZ_diff)
         # Must update beta first
-        beta_denomi = tf.math.maximum(tf.math.exp(logZ2_logZ1)*alpha1 -
-                                      tf.math.exp(logZ1_logZ0)*self.alpha,
-                                      1e-6)
-        self.beta.assign(self.beta/beta_denomi)
-        alpha_denomi = tf.math.maximum(tf.math.exp(logZ2_logZ1 - logZ1_logZ0) *
-                                       alpha1/alpha0  - 1.0,
-                                       1e-6)
-        self.alpha.assign(1.0/(alpha_denomi))
+        # Extract larger exponential
+        Pos_where = safe_exp(logZ2_logZ1)*(alpha1 - safe_exp(-logZ_diff)*self.alpha)
+        Neg_where = safe_exp(logZ1_logZ0)*(Z0Z2_Z1Z1*alpha1 - self.alpha)
+
+        beta_denomi = tf.where(logZ_diff >= 0, Pos_where, Neg_where)
+        self.beta.assign(safe_div(self.beta,
+                                  tf.maximum(beta_denomi,
+                                             tf.zeros_like(self.beta))))
+
+        alpha_denomi = Z0Z2_Z1Z1 * safe_div(alpha1, self.alpha) - 1.0
+        self.alpha.assign(safe_div(tf.constant(1.0,dtype=alpha_denomi.dtype),
+                                   tf.maximum(alpha_denomi,
+                                              tf.zeros_like(self.alpha))))
 
     def __call__(self,x):
         """
@@ -392,8 +428,8 @@ class PBP:
         for l in self.layers:
             x = l(x)
 
-        return x + (self.Normal.sample(x.shape) /
-                    (tf.math.sqrt(self.Gamma.sample(x.shape))+1e-6))
+        return x + safe_div(self.Normal.sample(x.shape),
+                            tf.math.sqrt(self.Gamma.sample(x.shape)))
 
 
     def predict(self,x):
@@ -415,7 +451,7 @@ class PBP:
         x = self._ensure_input(x)
         m, v = self._predict(x)
 
-        return m, v + self.beta/(self.alpha - 1)
+        return m, v + safe_div(self.beta, self.alpha - 1)
 
     @tf.function
     def _predict(self,x: tf.Tensor):
